from pydantic import BaseModel
from razorbill.crud import CRUD
from razorbill.router import Router
from razorbill.connectors.memory import MemoryConnector
from razorbill.connectors.alchemy.alchemy import AsyncSQLAlchemyConnector
from razorbill.connectors.mongo.mongo import AsyncMongoConnector
from sqlalchemy.orm import DeclarativeBase
from enum import Enum
from typing import Callable, Type, Any
from fastapi import Path, Depends


def builder_crud(
        url: str | None = None,
        schema: Type[BaseModel] | None = None,
        model: Type[DeclarativeBase] | None = None,
        memory_connector: bool | None = True,
        mongo_connector:bool | None = True,
        alchemy_connector: bool | None = False,
        pk: str | None = "id",
        session_maker: Any | None = None
):
    if alchemy_connector:
        if not url or not model:
            return None
        connector = AsyncSQLAlchemyConnector(
            model=model,
            url=url,
            pk_name=pk,
            session_maker=session_maker)
    elif mongo_connector:
        if not url or not schema:
            return None
        connector = AsyncMongoConnector(
            model=schema,
            url=url,
            pk_name=pk)

    else:
        if schema is None:
            return None

        connector = MemoryConnector(
            schema=schema,
            pk_name=pk
        )

    return CRUD(connector=connector)


def builder_router(
        url: str | None = None,
        schema: Type[BaseModel] | None = None,
        model: Type[DeclarativeBase] | None = None,
        memory_connector: bool | None = True,
        alchemy_connector: bool | None = False,
        mongo_connector: bool | None = False,
        create_schema: Type[BaseModel] | None = None,
        update_schema: Type[BaseModel] | None = None,
        overwrite_schema: bool = False,
        overwrite_create_schema: bool = False,
        overwrite_update_schema: bool = False,
        pk: str | None = "id",
        items_per_query: int = 10,
        item_name: str | None = None,
        parent_item_name: str | None = None,
        count_endpoint: bool | list[Callable] = True,
        get_all_endpoint: bool | list[Callable] = True,
        get_one_endpoint: bool | list[Callable] = True,
        create_one_endpoint: bool | list[Callable] = True,
        update_one_endpoint: bool | list[Callable] = True,
        delete_one_endpoint: bool | list[Callable] = True,
        path_item_parameter: Type[Path] | None = None,
        prefix: str = '',
        tags: list[str | Enum] | None = None,
        dependencies: list[Depends] | None = None,
        schema_slug: str | None = None,
        parent_schema: Type[BaseModel] | None = None,
        parent_model: Type[DeclarativeBase] | None = None,
        filters: list[str] | None = None,
        session_maker: Any | None = None,
        exclude_from_create_schema: list[str] = [],
        exclude_from_update_schema: list[str] = [],
):
    crud = builder_crud(
        url=url,
        schema=schema,
        model=model,
        memory_connector=memory_connector,
        alchemy_connector=alchemy_connector,
        mongo_connector=mongo_connector,
        pk=pk,
        session_maker=session_maker,
    )
    parent_crud = builder_crud(
        url=url,
        schema=parent_schema,
        model=parent_model,
        memory_connector=memory_connector,
        alchemy_connector=alchemy_connector,
        pk=parent_item_name,
        session_maker=session_maker,
    )
    router = Router(
        crud,
        items_per_query=items_per_query,
        item_name=item_name,
        parent_item_name=parent_item_name,
        count_endpoint=count_endpoint,
        get_all_endpoint=get_all_endpoint,
        get_one_endpoint=get_one_endpoint,
        create_one_endpoint=create_one_endpoint,
        update_one_endpoint=update_one_endpoint,
        delete_one_endpoint=delete_one_endpoint,
        parent_crud=parent_crud,
        path_item_parameter=path_item_parameter,
        prefix=prefix,
        tags=tags,
        dependencies=dependencies,
        schema_slug=schema_slug,
        create_schema=create_schema,
        update_schema=update_schema,
        overwrite_schema=overwrite_schema,
        overwrite_create_schema=overwrite_create_schema,
        overwrite_update_schema=overwrite_update_schema,
        filters=filters,
        exclude_from_create_schema=exclude_from_create_schema,
        exclude_from_update_schema=exclude_from_update_schema
    )
    return router



from enum import Enum
from fastapi import APIRouter, Path, Depends, Query, FastAPI
from typing import Optional
from razorbill.crud import CRUD
from razorbill.deps import (
    build_exists_dependency,
    build_last_parent_dependency,
    build_pagination_dependency,
    build_path_elements,
    init_deps,
    build_parent_populate_dependency,
    build_sorting_dependency,
)
from razorbill.exceptions import NotFoundError
from typing import Callable, Type
from pydantic import BaseModel
from razorbill.utils import (
    get_slug_schema_name,
    schema_factory,
    validate_filters,
    parent_schema_factory,
    create_schema_from_model_with_overwrite
)


def empty_dependency():
    return None


class Router(APIRouter):
    def __init__(
        self,
        crud: CRUD,
        items_per_query: int = 10,
        item_name: str | None = None,
        parent_item_name: str | None = None,
        count_endpoint: bool | list[Callable] = True,
        get_all_endpoint: bool | list[Callable] = True,
        get_one_endpoint: bool | list[Callable] = True,
        create_one_endpoint: bool | list[Callable] = True,
        update_one_endpoint: bool | list[Callable] = True,
        delete_one_endpoint: bool | list[Callable] = True,
        parent_crud: CRUD | None = None,
        path_item_parameter: Type[Path] | None = None,
        prefix: str = '',
        tags: list[str | Enum] | None = None,
        dependencies: list[Depends] | None = None,
        schema_slug: str | None = None,
        filters: list[str]|None = None,
        schema: Type[BaseModel] | None = None,
        create_schema: Type[BaseModel] | None = None,
        update_schema: Type[BaseModel] | None = None,
        overwrite_schema: bool = False,
        overwrite_create_schema: bool = False,
        overwrite_update_schema: bool = False,
        exclude_from_create_schema: list[str] = [],
        exclude_from_update_schema: list[str] = [],
        **kwargs
    ):
        self.parent_prefix = ''
        self.crud = crud
        self.parent_dependencies = dependencies
        self.pk = crud.connector.pk_name
        self.overwrite_schema = overwrite_schema
        self.overwrite_create_schema = overwrite_create_schema
        self.overwrite_update_schema = overwrite_update_schema
        self.Schema = schema if schema is not None else crud.connector.schema
        self.exclude_from_create_schema = exclude_from_create_schema
        self.exclude_from_update_schema = exclude_from_update_schema
        self.filters = validate_filters(self.Schema,
                                        filters) if filters is not None else None  # TODO нигде не используется
        if create_schema:
            self.create_schema = (
                create_schema
                if overwrite_create_schema
                else create_schema_from_model_with_overwrite(self.Schema, create_schema, pk_field_name=self.pk,
                                                             prefix="Create")
            )
        else:
            self.create_schema = schema_factory(self.Schema, self.exclude_from_create_schema + [self.pk])
        if update_schema:
            self.update_schema = (
                update_schema
                if overwrite_update_schema
                else create_schema_from_model_with_overwrite(self.Schema, update_schema, pk_field_name=self.pk,
                                                             prefix="")
            )
        else:
            self.update_schema = schema_factory(self.Schema, self.exclude_from_update_schema + [self.pk], prefix="Update")

        self.parent_item_name = parent_item_name
        self.CreateSchema = self.create_schema
        self.UpdateSchema = self.update_schema

        self._parent_id_dependency = Depends(empty_dependency)
        self._parent_populate_dependency = Depends(empty_dependency)

        if item_name is None:
            item_name = self.Schema.__name__

        if schema_slug is None:
            self._schema_slug = get_slug_schema_name(item_name)
        fields_to_exclude = ["id", "_id"]
        self._sort_field_dependency = build_sorting_dependency(self.Schema)
        if parent_crud is not None:
            if parent_item_name is None:
                self.parent_item_name = parent_crud.connector.schema.__name__
            parent_item_tag, _, parent_item_path = build_path_elements(self.parent_item_name)
            self.CreateSchema = schema_factory(self.create_schema, self.exclude_from_create_schema + [parent_item_tag])
            self.UpdateSchema = schema_factory(self.update_schema, self.exclude_from_update_schema + [parent_item_tag], prefix='Update')

            self.Schema = parent_schema_factory(self.Schema, self.parent_item_name)
            parent_exists_dependency = build_exists_dependency(parent_crud, parent_item_tag)
            self._parent_id_dependency = build_last_parent_dependency(parent_item_tag, crud.connector.type_pk)
            self._parent_populate_dependency = build_parent_populate_dependency()

            fields_to_exclude.append(parent_item_tag)
            if dependencies is not None:
                self.parent_dependencies.append(parent_exists_dependency)
            else:
                self.parent_dependencies = [parent_exists_dependency]

            if prefix is None:
                self.parent_prefix = parent_item_path
            else:
                self.parent_prefix += parent_item_path

        if tags is None:
            tags = [self._schema_slug]

        self.FilterSchema = schema_factory(self.CreateSchema, prefix='Filter', filters=filters)
        item_tag, path, item_path = build_path_elements(item_name)
        self._path = path
        #self._path_without_parent = path
        self._item_path = item_path
        self._path_field = Path(alias=item_tag) if path_item_parameter is None else path_item_parameter
        self._pagination_dependency = build_pagination_dependency(items_per_query)

        super().__init__(
            dependencies=dependencies,
            prefix=prefix,
            tags=tags,
            **kwargs
        )
        if count_endpoint:
            self._init_count_endpoint(self.parent_dependencies if self.parent_dependencies is not None else None)

        if get_all_endpoint:
            self._init_get_all_endpoint(self.parent_dependencies if self.parent_dependencies is not None else None)

        if get_one_endpoint:
            self._init_get_one_endpoint(get_one_endpoint)

        if create_one_endpoint:
            self._init_create_one_endpoint(self.parent_dependencies if self.parent_dependencies is not None else None)

        if update_one_endpoint:
            self._init_update_one_endpoint(update_one_endpoint)

        if delete_one_endpoint:
            self._init_delete_one_endpoint(delete_one_endpoint)

    def _init_count_endpoint(self, deps: list[Callable] | bool):
        @self.get(
            self.parent_prefix + self._path + "count", response_model=int, dependencies=init_deps(deps)
        )
        async def count(
                parent: dict[str, int] = self._parent_id_dependency,
                user_filter: self.FilterSchema = Depends(self.FilterSchema),
        ) -> int:
            payload = user_filter.dict(exclude_none=True)
            if parent is not None:
                payload |= parent
            return await self.crud.count(payload)

    def _init_get_all_endpoint(self, deps: list[Callable] | bool):

        @self.get(
            self.parent_prefix + self._path,
            response_model=list[self.Schema],
            dependencies=init_deps(deps),
            response_model_exclude_none=True,
            response_model_exclude_defaults=True,
            response_model_exclude_unset=True
        )
        async def get_many(
                pagination: tuple[str, int] = self._pagination_dependency,
                parent: dict[str, int] = self._parent_id_dependency,
                populate_parent: bool = self._parent_populate_dependency,
                user_filter: self.FilterSchema = Depends(self.FilterSchema),
                sorting: tuple[str, bool] = self._sort_field_dependency,
        ):
            payload = user_filter.dict(exclude_none=True)
            if parent is not None:
                payload |= parent
            skip, limit = pagination

            sort_field, sort_desc = sorting
            sorting = {sort_field: sort_desc} if sort_field else None
            items = await self.crud.get_many(skip, limit, filters=payload, populate=populate_parent, sorting=sorting)
            return items

    def _init_create_one_endpoint(self, deps: list[Callable] | bool):
        @self.post(

            self.parent_prefix + self._path,
            response_model=self.Schema,
            dependencies=init_deps(deps),
            response_model_exclude_none=True,
            response_model_exclude_defaults=True,
            response_model_exclude_unset=True
        )
        async def create_one(
                body: self.CreateSchema,
                parent: dict[str, int] = self._parent_id_dependency,
        ):
            payload = body.dict()
            if parent is not None:
                payload = body.dict() | parent
            item = await self.crud.create(payload)
            return item

    def _init_get_one_endpoint(self, deps: list[Callable] | bool):
        @self.get(
            self._item_path,
            response_model=self.Schema,
            dependencies=init_deps(deps),
            response_model_exclude_none=True,
            response_model_exclude_defaults=True,
            response_model_exclude_unset=True
        )
        async def get_one(
                item_id: int | str = self._path_field,
                #parent: dict[str, int] = self._parent_id_dependency,
                populate_parent: bool | str = self._parent_populate_dependency,
        ):
            #item = await self.crud.get_one(item_id, parent, populate=populate_parent)
            item = await self.crud.get_one(item_id, populate=populate_parent)
            if item:
                return item
            raise NotFoundError(self.Schema.__name__, self._path_field.alias, item_id)


    def _init_update_one_endpoint(self, deps: list[Callable] | bool):
        @self.put(
            self._item_path,
            response_model=self.Schema,
            dependencies=init_deps(deps),
            response_model_exclude_none=True,
            response_model_exclude_defaults=True,
            response_model_exclude_unset=True
        )
        async def update_one(
                *,
                #parent: dict[str, int] = self._parent_id_dependency,
                item_id: int | str = self._path_field,
                body: self.UpdateSchema,
        ):
            payload = body.dict(exclude_unset=True)
            # if parent is not None:
            #     payload = body.dict(exclude_unset=True) | parent

            item = await self.crud.update(item_id, payload)  # type: ignore
            if item:
                return item
            raise NotFoundError(self.Schema.__name__, self._path_field.alias, item_id)

    def _init_delete_one_endpoint(self, deps: list[Callable] | bool):
        @self.delete(self._item_path, dependencies=init_deps(deps))
        async def delete_one(
                #parent: dict[str, int] = self._parent_id_dependency,
                item_id: int | str = self._path_field,
        ):
            item = await self.crud.delete(item_id)
            print(item)
            if item:
                return item
            raise NotFoundError(self.Schema.__name__, self._path_field.alias, item_id)



from abc import ABC
from typing import Any, Type
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from pydantic import BaseModel, validate_arguments
from pymongo.errors import DuplicateKeyError
import pymongo
from bson import ObjectId
from beanie import PydanticObjectId
from razorbill.connectors.base import BaseConnector
from razorbill.connectors.mongo.utils import _prepare_result, validate_id
from beanie import init_beanie, Document
from functools import wraps
from razorbill.connectors.mongo.utils import create_beanie_model


class AsyncMongoConnectorException(Exception):
    pass


def ensure_initialized(func):
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        if not self.initialized:
            await self.init_beanie()
        return await func(self, *args, **kwargs)

    return wrapper


class AsyncMongoConnector(BaseConnector):
    @validate_arguments
    def __init__(self, url: str, model: Type[BaseModel], pk_name: str = "id") -> None:
        self.document_schema = create_beanie_model(model, model.__name__, pk_name)
        self.db_name = url.split('/')[-1]
        self.client: AsyncIOMotorClient = AsyncIOMotorClient(url)
        self._schema = self.document_schema  # update_mongo_schema(model)
        self._pk_name = pk_name
        self.initialized = False

    @property
    def pk_name(self) -> str:
        return self._pk_name

    @property
    def schema(self) -> Type[BaseModel]:
        return self._schema

    @property
    def type_pk(self) -> Type[str]:
        return str

    async def init_beanie(self):
        await init_beanie(database=self.client[self.db_name], document_models=[self.document_schema])
        self.initialized = True

    @ensure_initialized
    async def count(self, filters: dict[str, Any] = {}) -> int:
        if filters:
            query = self.document_schema.find(filters)
        else:
            query = self.document_schema.find()
        return await query.count()

    @ensure_initialized
    async def create_one(self, obj: dict[str, Any]) -> dict[str, Any]:
        try:
            new_document = self.document_schema(**obj)
            new_obj = await new_document.insert()
            return new_obj

        except DuplicateKeyError:
            raise AsyncMongoConnectorException(f"Duplicate key error")

    @ensure_initialized
    async def get_many(
            self,
            skip: int,
            limit: int,
            filters: dict[str, Any] = {},
            populate: bool = False,
            sorting: dict[str, bool] = None
    ) -> list[dict[str, Any]]:
        query = self.document_schema.find(filters) if filters else self.document_schema.find()
        if sorting:
            sort_fields = [(field if field != 'id' else "_id", pymongo.DESCENDING if sort_desc else pymongo.ASCENDING)
                           for field, sort_desc in sorting.items()]
            query = query.sort(sort_fields)
        query = query.skip(skip).limit(limit)
        return await query.to_list()

    @ensure_initialized
    async def get_one(
            self,
            obj_id: str | int,
            filters: dict[str, Any] = {},
            populate: bool = False
    ) -> dict[str, Any] | None:
        obj_id = validate_id(obj_id)
        if obj_id is None:
            return None
        result = await self.document_schema.get(obj_id)
        if not result:
            return None

        return _prepare_result(result, [])

    @ensure_initialized
    async def update_one(
            self, obj_id: str | int,
            obj: dict[str, Any],
            filters: dict[str, Any] = {}
    ) -> dict[str, Any]:

        document = await self.document_schema.get(obj_id)
        result = await document.update({"$set": obj})

        if not result:
            return None

        return result

    @ensure_initialized
    async def delete_one(self, obj_id: str | int, filters: dict[str, Any] = {}) -> bool:
        document = await self.document_schema.get(obj_id)
        result = await document.delete()

        if not result:
            return None

        return result

#
# class AsyncMongoConnector(BaseConnector):
#     @validate_arguments
#     def __init__(self, url: str, model: Type[BaseModel], pk_name: str = "id") -> None:
#         self.document_schema = create_beanie_model(model, model.__name__, pk_name)
#
#         self.client: AsyncIOMotorClient = AsyncIOMotorClient(url)
#         self.db: AsyncIOMotorDatabase = self.client['razorbill']
#         self.collection = self.db[self.document_schema.__name__.lower()]
#         self._schema = update_mongo_schema(model)
#         self._pk_name = pk_name
#
#
#     @property
#     def pk_name(self) -> str:
#         return self._pk_name
#
#     @property
#     def schema(self) -> Type[BaseModel]:
#         return self._schema
#
#     # async def init_beanie(self):
#     #     await init_beanie(database=self.client.db_name, document_schemas=[self.document_schema])
#
#     async def create_one(self, obj: dict[str, Any]) -> dict[str, Any]:
#         try:
#             result = await self.collection.insert_one(obj)
#             obj['_id'] = result.inserted_id
#             return obj
#
#         except DuplicateKeyError:
#             raise AsyncMongoConnectorException(f"Duplicate key error")
#
#     async def count(self, filters: dict[str, Any] = {}) -> int:
#
#         return await self.collection.count_documents(filters if filters else {})
#
#     async def get_many(
#             self,
#             skip: int,
#             limit: int,
#             filters: dict[str, Any] = {},
#             populate: bool = False,
#             sorting: dict[str, bool] = None
#     ) -> list[dict[str, Any]]:
#         cursor = self.collection.find(filters).skip(skip).limit(limit)
#
#         if sorting:
#             sort = [(field, pymongo.DESCENDING if sort_desc else pymongo.ASCENDING) for field, sort_desc in
#                     sorting.items()]
#             cursor = cursor.sort(sort)
#
#         results = await cursor.to_list(length=limit)
#         print(results)
#         return results
#         #return [_prepare_result(result) for result in results]
#
#     async def get_one(
#             self,
#             obj_id: str | int,
#             filters: dict[str, Any] = {},
#             populate: bool = False
#     ) -> dict[str, Any] | None:
#         query = {'_id': obj_id}
#         result = await self.collection.find_one(query, filters)
#         print(result)
#         if not result:
#             return None
#
#         return _prepare_result(result, [])
#
#     async def update_one(
#             self, obj_id: str | int,
#             obj: dict[str, Any],
#             filters: dict[str, Any] = {}
#     ) -> dict[str, Any]:
#         query = {'_id': obj_id}
#         update = {'$set': obj}
#
#         result = await self.collection.find_one_and_update(query, update, return_document=pymongo.ReturnDocument.AFTER)
#
#         if not result:
#             return None
#
#         return _prepare_result(result)
#
#     async def delete_one(self, obj_id: str | int, filters: dict[str, Any] = {}) -> bool:
#         query = {'_id': obj_id}
#         result = await self.collection.delete_one(query)
#         return result.deleted_count > 0



from abc import ABC, abstractmethod
from typing import Any, Type
from pydantic import BaseModel


class BaseConnector(ABC):
    @property
    @abstractmethod
    def schema(self) -> Type[BaseModel]:
        pass

    @property
    @abstractmethod
    def pk_name(self) -> str:
        pass

    @abstractmethod
    def type_pk(self) -> Type[Any]:
        pass

    @abstractmethod
    async def count(self, filters: dict[str, Any] | None = None) -> int:
        pass

    @abstractmethod
    async def get_many(
            self, skip: int, limit: int, filters: dict[str, Any] | None = None, populate: bool = False,
            sorting: dict[str, bool] | None = None
    ) -> list[dict[str, Any]]:
        pass

    @abstractmethod
    async def get_one(self, obj_id: str | int, populate: bool | str = False) -> dict[str, Any]:
        pass

    @abstractmethod
    async def create_one(self, obj: dict[str, Any]) -> dict[str, Any]:
        pass

    @abstractmethod
    async def update_one(
            self, obj_id: str | int, obj: dict[str, Any]) -> dict[str, Any]:
        pass

    @abstractmethod
    async def delete_one(self, obj_id: str | int) -> dict[str, Any] | None:
        pass



from typing import Any, Type, Container, Optional
from sqlalchemy import inspect
from sqlalchemy.orm import class_mapper, DeclarativeBase, ColumnProperty
from sqlalchemy.ext.declarative import DeclarativeMeta
from pydantic import BaseModel, create_model, BaseConfig


class OrmConfig(BaseConfig):
    orm_mode = True


def _sqlalchemy_to_pydantic(
        db_model: Type,
        *,
        config: Type = OrmConfig,
        exclude: Container[str] = [],
        prefix: str | None = None,
        base_pydantic_model: Type[BaseModel] | None = None,
) -> Type[BaseModel]:
    model_name = db_model.__name__

    if prefix is not None:
        model_name = prefix + model_name

    mapper = inspect(db_model)
    fields = {}
    for attr in mapper.attrs:
        if isinstance(attr, ColumnProperty):
            if attr.columns:
                name = attr.key
                if name in exclude:
                    continue
                column = attr.columns[0]
                python_type: Optional[type] = None
                if hasattr(column.type, "impl"):
                    if hasattr(column.type.impl, "python_type"):
                        python_type = column.type.impl.python_type
                elif hasattr(column.type, "python_type"):
                    python_type = column.type.python_type
                assert python_type, f"Could not infer python_type for {column}"
                default = None
                if column.default is None and not column.nullable:
                    default = ...
                fields[name] = (python_type, default)
    pydantic_model = create_model(
        model_name, __base__=base_pydantic_model, __config__=config, **fields
    )
    return pydantic_model


def _pydantic_to_sqlalchemy(pydantic_obj: BaseModel, sqlalchemy_model: DeclarativeMeta) -> DeclarativeBase:
    data = pydantic_obj.dict()
    mapped_fields = {}
    mapper = class_mapper(sqlalchemy_model)

    for field in mapper.columns:
        field_name = field.key
        if field_name in data:
            mapped_fields[field_name] = data[field_name]

    sqlalchemy_instance = sqlalchemy_model(**mapped_fields)
    return sqlalchemy_instance


def _prepare_result(item, parent_relationships):
    if item:

        schema_dict = item.__dict__
        result_dict = _object_to_dict(item)
        for parent_model_name in parent_relationships:
            parent_data = schema_dict.pop(parent_model_name)
            parent_dict = _object_to_dict(parent_data)
            result_dict[parent_model_name] = parent_dict
        return result_dict
    return None


def _get_parent_relationships(model: Type[DeclarativeBase], parent_name: [str]) -> list[str]:
    parent_relationships = []
    for column in model.__table__.columns:
        for fk in column.foreign_keys:
            for key in parent_name:
                if key == column.name:
                    for rel in model.__mapper__.relationships:
                        if fk.column.table == rel.entity.class_.__table__:
                            parent_relationships.append(rel.key)
    return parent_relationships


def _object_to_dict(obj: Type[DeclarativeBase]) -> dict[str, Any]:
    mapper = class_mapper(obj.__class__)
    columns = [column.key for column in mapper.columns]
    return {column: getattr(obj, column) for column in columns}


import re
from razorbill._types import T
from typing import Any

from typing import Type, TypeVar
from pydantic import BaseModel, create_model, Field, BaseConfig

T = TypeVar("T", bound=BaseModel)


class OrmConfig(BaseConfig):
    orm_mode = True


def create_schema_from_model_with_overwrite(
    schema_cls: Type[T], 
    overwrite_cls: Type[BaseModel],
    pk_field_name: str = "_id", 
    prefix: str = "Create"
) -> Type[T]:
    fields = {}
    for field_name, field_info in schema_cls.__fields__.items():
        if field_name != pk_field_name:
            for overwrite_field_name, overwrite_field_info in overwrite_cls.__fields__.items():
                if field_name == overwrite_field_name:
                    fields[field_name] = (overwrite_field_info.annotation, ... if overwrite_field_info.is_required() else None)
                    break
            fields[field_name] = (field_info.annotation, ... if field_info.is_required() else None)

    name = prefix + schema_cls.__name__
    schema: Type[T] = create_model(__model_name=name, __base__=overwrite_cls, **fields)  # type: ignore
    return schema


def parent_schema_factory(schema_cls: Type[T], pk_field_name: str) -> Type[T]:
    fields = {
        name: (info.annotation, ... if info.is_required() else None)
        for name, info in schema_cls.__fields__.items()
    }
    fields[pk_field_name] = (dict[str, Any], Field(None))
    name = schema_cls.__name__
    config = getattr(schema_cls, "__config__", None)
    schema: Type[T] = create_model(__model_name=name, **fields, __config__=config)  # type: ignore
    return schema


def schema_factory(
    schema_cls: Type[T], 
    exclude_fields: list[str] = ["_id"], 
    prefix: str = "Create", 
    filters: list[str]|None = None
) -> Type[T]:
    if filters is None:
        if prefix == 'Filter':
            fields = {}
        else:
            fields = {
                name: (info.annotation, ... if info.is_required() else None)
                for name, info in schema_cls.__fields__.items()
                if name not in exclude_fields
            }
    else:
        fields = {
            name: (info.annotation, Field(None))
            for name, info in schema_cls.__fields__.items()
            if name not in exclude_fields and name in filters
        }

    name = prefix + schema_cls.__name__
    schema: Type[T] = create_model(__model_name=name, __base__=None, __config__=OrmConfig, **fields)  # type: ignore
    return schema


def get_slug_schema_name(schema_name: str) -> str:
    chunks = re.findall("[A-Z][^A-Z]*", schema_name)
    return "_".join(chunks).lower()


def validate_filters(
        schema_cls: Type[T],
        filters: list[str]
):
    valid_filters = [
        filter_field for filter_field 
        in filters if filter_field 
        in schema_cls.__annotations__
    ]
    return valid_filters



from typing import Any, Type, Container, Optional
from pydantic.fields import Field, ModelField
from pydantic import BaseModel, create_model, BaseConfig
from beanie import Document
from bson import ObjectId
from bson.errors import InvalidId
from beanie import PydanticObjectId
from pydantic import ValidationError
from beanie.odm.documents import DocType
from beanie.odm.utils.pydantic import get_model_fields


def create_beanie_model(
        pydantic_model: Type[BaseModel],
        db_model_name: str,
        exclude: Container[str] = (),
) -> Type[Document]:
    fields = {
        name: (info.annotation, ...)
        for name, info in pydantic_model.__fields__.items()
        if name not in exclude
    }
    beanie_model = create_model(
        db_model_name, __base__=Document, __config__=None, **fields
    )
    return beanie_model


def update_mongo_schema(
        schema_cls: Type[BaseModel],
        pk_field_name: str = "id"
) -> BaseModel:

    fields = {
        f.name: (f.type_, ...)
        for f in schema_cls.__fields__.values()
    }
    fields['id'] = ModelField(
        name='id',
        type_=Optional[str],
        required=False,
        default=None,
        alias='_id',
        class_validators=None,
        model_config=BaseConfig,

    )

    name = schema_cls.__name__
    schema: Type[T] = create_model(__model_name=name, **fields)  # type: ignore
    return schema


def _prepare_result(item, parent_relationships):
    if item:
        schema_dict = item
        for parent_model_name in parent_relationships:
            parent_data = schema_dict.pop(parent_model_name, None)
            if parent_data is not None:
                schema_dict[parent_model_name] = parent_data
        return schema_dict
    return None


def _get_parent_relationships(model_name, filters, collection):
    parent_relationships = []
    for key, value in filters.items():
        # Поиск по внешним ключам или другим полям, связанным с родительской коллекцией
        for parent in collection.find({key: value}):
            for parent_field in parent:
                parent_relationships.append(parent_field)
    return parent_relationships


def validate_id(id_str: str) -> PydanticObjectId:
    try:
        obj_id = ObjectId(id_str)
        pydantic_obj_id = PydanticObjectId(obj_id)
        return pydantic_obj_id
    except (ValueError, ValidationError, InvalidId):
        return None


from fastapi import HTTPException, Depends, Path, Request, Query
from razorbill.crud import CRUD
import re
from typing import Callable, Type
from pydantic import BaseModel, Field
from razorbill.exceptions import NotFoundError


def build_exists_dependency(
        crud: CRUD, item_tag: str
) -> Depends:
    """Зависимость, которая берет id элемента из URL и проверяет есть ли данный элемент в базе"""

    async def dep(
            item_id: int | str = Path(alias=item_tag),
    ):
        item = await crud.get_one(item_id)
        if item is None:
            raise NotFoundError(crud.connector.schema.__name__, item_tag, item_id)

    return dep


# TODO нужно ли тут возвращать словарь, надо подумать
def build_last_parent_dependency(item_tag: str, type_pk: Type[str | int]) -> Depends:
    """Зависимость, которая передает в endpoint id родителя (если он есть)"""

    async def dep(request: Request):
        item_id = request.path_params.get(item_tag)
        if item_id is None:
            return {}
        return {item_tag: type_pk(item_id)}

    return Depends(dep)


def build_parent_populate_dependency() -> Depends:
    async def dep(
            populate_parent: bool | str = Query(False),
    ):
        return populate_parent

    return Depends(dep)


def build_sorting_dependency(obj: Type[BaseModel]) -> Depends:
    def get_sortable_fields():
        return list(obj.__fields__.keys())

    async def dep(
            sort_field: str = Query(None, description="Field to sort by", enum=get_sortable_fields()),
            sort_desc: bool = Query(None)
    ):
        return sort_field, sort_desc

    return Depends(dep)


def build_pagination_dependency(max_limit: int | None = None) -> Depends:
    """Зависимость, которая валидируют пагинационный параметры запроса"""

    def create_query_validation_exception(field: str, msg: str) -> HTTPException:
        return HTTPException(
            422,
            detail={
                "detail": [
                    {"loc": ["query", field], "msg": msg, "type": "type_error.integer"}
                ]
            },
        )

    def pagination(skip: int = 0, limit: int | None = max_limit) -> tuple[int, int | None]:
        if skip < 0:
            raise create_query_validation_exception(
                field="skip",
                msg="skip query parameter must be greater or equal to zero",
            )

        if limit is not None:
            if limit <= 0:
                raise create_query_validation_exception(
                    field="limit", msg="limit query parameter must be greater then zero"
                )

            elif max_limit and max_limit < limit:
                raise create_query_validation_exception(
                    field="limit",
                    msg=f"limit query parameter must be less then {max_limit}",
                )

        return skip, limit

    return Depends(pagination)


def init_deps(funcs: list[Callable] | bool) -> list[Depends]:
    if isinstance(funcs, list):
        return [Depends(func) for func in funcs]
    return []


def camel_to_snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def build_path_elements(name: str) -> tuple[str, str, str]:
    """Создает строковые элементы URL"""
    name = camel_to_snake(name)
    item_tag = name + "_id"
    item_path_tag = "{" + item_tag + "}"
    path = f"/{name}/"
    item_path = path + item_path_tag

    return item_tag, path, item_path



from fastapi import HTTPException

class AppError(Exception):
    def __init__(self, message: str, status_code: int):
        self.status_code = status_code
        self.message = message


class NotFoundError(AppError):
    def __init__(self, model_name: str, field_name: str, field_value: int):
        self.status_code = 404
        self.message = f"{model_name} with {field_name}={field_value} not found"
        raise HTTPException(status_code=self.status_code, detail=self.message)


from typing import Any, Type
from .base import BaseConnector
from pydantic import BaseModel

_inmemory_storage: dict[
    str, dict[int, dict[str, Any]]] = {}  # key = schema_name, value = {key = (pk, parent_pk | None), value = value }


class MemoryConnector(BaseConnector):
    def __init__(self, schema: Type[BaseModel], pk_name: str = "id") -> None:
        self._id = 1
        self._pk_name = pk_name
        self._schema = schema
        _inmemory_storage[schema.__name__] = {}

    @property
    def pk_name(self) -> str:
        return self._pk_name

    @property
    def schema(self) -> Type[BaseModel]:
        return self._schema

    def _get_next_id(self) -> int:
        id_ = self._id
        self._id += 1

        return id_

    async def create_one(self, obj: dict[str, Any]) -> dict[str, Any]:
        id = self._get_next_id()
        obj_with_id = {**obj, 'id': id}
        _inmemory_storage[self.schema.__name__][id] = obj_with_id
        return obj_with_id

    async def count(self, filters: dict[str, Any] | None = None) -> int:
        filtered_models = _inmemory_storage.get(self._schema.__name__, {})
        if filters:
            filtered_models = [obj for obj in filtered_models.values() if
                               all(obj.get(key) == value for key, value in filters.items())]
        return len(filtered_models)

    async def get_one(self, obj_id: str | int, filters: dict[str, Any] | None = None, populate: list[str] = None, ) -> \
    dict[str, Any] | None:
        obj = _inmemory_storage[self._schema.__name__].get(obj_id)
        foreign_keys = {}
        if obj is not None:
            if filters:
                matches_filters = all(obj.get(key) == value for key, value in filters.items())
                if not matches_filters:
                    return None

            if populate is not None:
                for field in self._schema.__fields__:
                    for fk in populate:
                        parent_fk = fk.replace('Schema', '').lower() + '_id'
                        parent_relationship = fk.replace('Schema', '').lower()
                        if field == parent_fk:
                            foreign_keys[fk] = {
                                'parent_fk': fk.replace('Schema', '').lower() + '_id',
                                'parent_relationship': parent_relationship
                            }

            for schema_name, parent in foreign_keys.items():
                fk_value = obj.get(parent['parent_fk'])
                parent_obj = _inmemory_storage[schema_name].get(fk_value)
                if parent:
                    del obj[parent['parent_fk']]
                    obj[parent['parent_relationship']] = parent_obj

            return obj
        return None

    async def get_many(
            self,
            skip: int,
            limit: int,
            filters: dict[str, Any] | None = None,
            populate: list[str] | None = None,
    ) -> list[dict[str, Any]]:
        result = []
        foreign_keys = {}

        for obj_id, obj in _inmemory_storage.get(self._schema.__name__, {}).items():
            matches_filters = True

            if filters:
                for key, value in filters.items():
                    if obj.get(key) != value:
                        matches_filters = False
                        break

            if matches_filters:
                if populate is not None:
                    for field in self._schema.__fields__:
                        for fk in populate:
                            parent_fk = fk.replace('Schema','').lower() + '_id'
                            parent_relationship = fk.replace('Schema','').lower()
                            if field == parent_fk:
                                foreign_keys[fk] = {
                                    'parent_fk' :fk.replace('Schema','').lower() + '_id',
                                    'parent_relationship': parent_relationship
                                }

                for schema_name, parent in foreign_keys.items():
                    fk_value = obj.get(parent['parent_fk'])
                    parent_obj = _inmemory_storage[schema_name].get(fk_value)
                    if parent:
                        del obj[parent['parent_fk']]
                        obj[parent['parent_relationship']] = parent_obj
                result.append(obj)


        return result[skip: skip + limit]

    async def update_one(
            self, obj_id: str | int,
            obj: dict[str, Any],
            filters: dict[str, Any] | None = None
    ) -> dict[str, Any] | None:
        if obj_id is not None:
            update_obj = _inmemory_storage.get(self._schema.__name__, {}).get(obj_id)
            if update_obj is not None:
                if filters:
                    matches_filters = all(update_obj.get(key) == value for key, value in filters.items())
                    if not matches_filters:
                        return None
                _obj = obj.copy()
                updated_fields = {key: value for key, value in _obj.items() if key != "id"}
                if updated_fields:
                    updated_fields["id"] = obj_id
                    _inmemory_storage[self._schema.__name__][obj_id].update(updated_fields)
                    return _inmemory_storage[self._schema.__name__][obj_id]
                else:
                    return update_obj
        return None

    async def delete_one(self, obj_id: str | int, filters: dict[str, Any] | None = None) -> bool:
        obj = _inmemory_storage[self._schema.__name__].get(obj_id)
        if obj is not None:
            if filters:
                matches_filters = all(obj.get(key) == value for key, value in filters.items())
                if not matches_filters:
                    return False
            del _inmemory_storage[self._schema.__name__][obj_id]
            return True
        return False



from typing import Any, Type, Container, Optional
from sqlalchemy import and_, func, update, inspect, desc
import sqlalchemy
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import class_mapper, joinedload, DeclarativeBase, sessionmaker
from sqlalchemy.future import select
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from pydantic import BaseModel, create_model
from pydantic import validate_arguments

from razorbill.connectors.base import BaseConnector
from razorbill.connectors.alchemy.utils import _sqlalchemy_to_pydantic, _prepare_result, _get_parent_relationships, \
    _object_to_dict


class AsyncSQLAlchemyConnectorException(Exception):
    pass


class AsyncSQLAlchemyConnector(BaseConnector):
    @validate_arguments
    def __init__(self, url: str, model: Type[DeclarativeBase], session_maker: Any = None,
                 pk_name: str = "id", **kwargs) -> None:
        self.model = model
        if session_maker is None:
            self.engine = create_async_engine(url, **kwargs)
            self.session_maker = sessionmaker(self.engine, class_=AsyncSession, expire_on_commit=False)
        else:
            self.session_maker = session_maker
        self._schema = _sqlalchemy_to_pydantic(self.model)
        self._pk_name = pk_name

    @property
    def schema(self) -> Type[BaseModel]:
        return self._schema

    @property
    def pk_name(self) -> str:
        return self._pk_name

    @property
    def type_pk(self) -> Type[int]:
        return int

    async def create_one(
            self, obj: dict[str, Any]
    ) -> dict[str, Any]:
        sql_model = self.model(**obj)
        async with self.session_maker.begin() as session:
            session.add(sql_model)
            try:
                await session.commit()
                created_sql_model = await session.merge(sql_model)
                return _object_to_dict(created_sql_model)
            except sqlalchemy.exc.IntegrityError as error:
                raise AsyncSQLAlchemyConnectorException(f"Some of relations objects does not exists")

    async def count(
            self, filters: dict[str, Any] = {}
    ) -> int:
        where = []
        if filters is not None:
            where = [getattr(self.model, key) == value for key, value in filters.items()]

        statement = select(func.count()).select_from(
            select(self.model).where(and_(True, *where)).subquery()
        )
        async with self.session_maker.begin() as session:
            count = await session.scalar(statement)
        return count

    async def get_many(
            self,
            skip: int,
            limit: int,
            filters: dict[str, Any] = {},
            populate: bool = False,
            sorting: dict[str, bool] = None
    ) -> list[dict[str, Any]]:
        statement = select(self.model)

        parent_relationships = []
        where = []
        if filters:
            where = [getattr(self.model, key) == value for key, value in filters.items()]
        if populate:
            parent_relationships = _get_parent_relationships(self.model, filters.keys())
            relationship_attrs = [getattr(self.model, field) for field in parent_relationships]
            statement = statement.options(
                *[joinedload(attr) for attr in relationship_attrs]
            )
            relationship_attrs = [getattr(self.model, field) for field in parent_relationships]
            statement = statement.where(and_(True, *where)).options(
                *[joinedload(attr) for attr in relationship_attrs]
            ).offset(skip).limit(limit)

        else:
            statement = statement.where(and_(True, *where)).offset(skip).limit(limit)

        if sorting:
            for field, sort_desc in sorting.items():
                sort_column = getattr(self.model, field)
                if sort_desc:
                    statement = statement.order_by(desc(sort_column))
                else:
                    statement = statement.order_by(sort_column)

        async with self.session_maker.begin() as session:
            result = await session.execute(statement)
            items = result.scalars().all()

        return [_prepare_result(item, parent_relationships) for item in items]

    async def get_one(
            self,
            obj_id: str | int,
            populate: bool | str = False,
    ) -> dict[str, Any] | None:
        statement = select(self.model)
        parent_relationships = []
        statement = statement.where(self.model.id == int(obj_id))

        if populate:
            parent_relationships = _get_parent_relationships(self.model, [populate])
            relationship_attrs = [getattr(self.model, field) for field in parent_relationships]
            statement = statement.options(
                *[joinedload(attr) for attr in relationship_attrs]
            )
            relationship_attrs = [getattr(self.model, field) for field in parent_relationships]
            statement = statement.options(
                *[joinedload(attr) for attr in relationship_attrs]
            )
        async with self.session_maker.begin() as session:
            query = await session.execute(statement)
            try:
                item = query.scalars().one_or_none()
            except NoResultFound:
                item = None

        return _prepare_result(item, parent_relationships) if item else None

    async def update_one(
            self, obj_id: str | int,
            obj: dict[str, Any]
    ) -> dict[str, Any]:
        statement = (
            update(self.model)
            .values(obj)
            .where(self.model.id == int(obj_id))
            .execution_options(synchronize_session="fetch")
        )

        try:
            async with self.session_maker.begin() as session:
                await session.execute(statement)
                await session.commit()
                updated_obj = await self.get_one(obj_id)

            return updated_obj if updated_obj else None
        except sqlalchemy.exc.IntegrityError as error:
            raise AsyncSQLAlchemyConnectorException(f"Some of relations objects does not exists: {error}")

    async def delete_one(self, obj_id: str | int) -> dict[str, Any] | None:
        async with self.session_maker.begin() as session:
            statement = select(self.model).where(self.model.id == int(obj_id))
            where = []
            statement = statement.where(and_(True, *where))

            query = await session.execute(statement)
            item = query.scalars().one_or_none()

            if item is not None:
                await session.delete(item)
                await session.commit()
                return _object_to_dict(item)
            return None



from typing import Any, Type, Callable
from pydantic import validate_arguments
from razorbill.connectors.base import BaseConnector

class CRUD:
    def __init__(
            self,
            connector: Type[BaseConnector] | None = None,
    ):

        self._connector = connector

        self._before_create_func = None
        self._before_update_func = None
        self._before_delete_func = None
        self._before_get_one_func = None
        self._before_get_many_func = None

        self._after_create_func = None
        self._after_update_func = None
        self._after_delete_func = None
        self._after_get_one_func = None
        self._after_get_many_func = None

    @property
    def connector(self) -> Type[BaseConnector]:
        return self._connector

    @validate_arguments
    def before_create(self, func: Callable) -> Callable:
        self._before_create_func = func
        return func

    @validate_arguments
    def before_update(self, func: Callable) -> Callable:
        self._before_update_func = func
        return func

    @validate_arguments
    def before_delete(self, func: Callable) -> Callable:
        self._before_delete_func = func
        return func

    @validate_arguments
    def before_get_one(self, func: Callable) -> Callable:
        self._before_get_one_func = func
        return func

    @validate_arguments
    def before_get_many(self, func: Callable) -> Callable:
        self._before_get_many_func = func
        return func

    @validate_arguments
    def after_create(self, func: Callable) -> Callable:
        self._after_create_func = func
        return func

    @validate_arguments
    def after_update(self, func: Callable) -> Callable:
        self._after_update_func = func
        return func

    @validate_arguments
    def after_delete(self, func: Callable) -> Callable:
        self._after_delete_func = func
        return func

    @validate_arguments
    def after_get_one(self, func: Callable) -> Callable:
        self._after_get_one_func = func
        return func

    @validate_arguments
    def after_get_many(self, func: Callable) -> Callable:
        self._after_get_meny_func = func
        return func

    async def count(self, filters: dict[str, Any] = {}) -> int:
        return await self._connector.count(filters=filters)

    async def get_one(self, obj_id: str | int,  populate: bool = False) -> dict[
        str, Any]:
        if self._before_get_one_func is not None:
            await self._before_get_one_func(obj_id, populate)
        item = await self._connector.get_one(obj_id=obj_id, populate=populate)
        if self._after_get_one_func is not None:
            item = await self._after_get_one_func(item)
        return item

    async def get_many(self, skip: int, limit: int, filters: dict[str, Any] = {},
                       populate: bool = False,
                       sorting: dict[str, bool] = {}) -> list[dict[str, Any]]:
        _obj = None
        if self._before_get_many_func is not None:
            await self._before_get_many_func(skip, limit, filters, sorting, populate)

        items = await self._connector.get_many(skip=skip, limit=limit,
                                               filters=filters, populate=populate, sorting=sorting)
        if self._after_get_many_func is not None:
            items = await self._after_get_many_func(items)
        return items

    async def create(self, obj: Type[dict[str, Any]]) -> dict[str, Any]:
        _obj = None
        if self._before_create_func is not None:
            _obj = await self._before_create_func(obj)
        if _obj is None:
            _obj = obj
        record = await self._connector.create_one(obj=_obj)
        if self._after_create_func is not None:
            modified_record = await self._after_create_func(record)
            if modified_record is not None:
                record = modified_record
        return record

    async def update(self, obj_id: str | int, obj: Type[dict[str, Any]]) -> dict[
        str, Any]:
        _obj = None
        if self._before_update_func is not None:
            _obj = await self._before_update_func(obj_id, obj)
        if _obj is None:
            _obj = obj
        record = await self._connector.update_one(obj_id=obj_id, obj=_obj)
        if self._after_update_func is not None:
            modified_record = await self._after_update_func(record)
            if modified_record is not None:
                record = modified_record
        return record

    async def delete(self, obj_id: str | int):
        if self._before_delete_func is not None:
            await self._before_delete_func(obj_id)
        record = await self._connector.delete_one(obj_id=obj_id)
        if self._after_delete_func is not None:
            await self._after_delete_func(record)
        return record



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['razorbill',
 'razorbill.connectors',
 'razorbill.connectors.alchemy',
 'razorbill.connectors.mongo']

package_data = \
{'': ['*']}

install_requires = \
['alembic>=1.12.0,<2.0.0',
 'asyncpg>=0.28.0,<0.29.0',
 'asyncpgsa>=0.27.1,<0.28.0',
 'beanie>=1.21.0,<2.0.0',
 'fastapi>=0.103.0,<0.104.0',
 'greenlet>=2.0.2,<3.0.0',
 'loguru>=0.7.0,<0.8.0',
 'motor>=3.3.1,<4.0.0',
 'psycopg2-binary>=2.9.7,<3.0.0',
 'pytest-asyncio>=0.21.1,<0.22.0',
 'pytest>=7.4.0,<8.0.0',
 'sqlalchemy>=2.0.17,<3.0.0',
 'uvicorn>=0.23.2,<0.24.0']

setup_kwargs = {
    'name': 'razorbill',
    'version': '0.1.19',
    'description': 'Razorbill is a framework for quickly creating API applications using only the description of the data model.',
    'long_description': 'Razorbill is a framework for quickly creating API applications using only the description of the data model.',
    'author': 'Nikita Irgashev',
    'author_email': 'nik.irg@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
